/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick SIMIAN

	OneTrick SIMIAN is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick SIMIAN is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick SIMIAN.  If not, see <http://www.gnu.org/licenses/>.
*/

module punklabs.ui.knobbase;

import gui;
import core.atomic;
import dplug.core;
import dplug.gui;
import dplug.canvas;
import dplug.client;

import std.algorithm;
import std.math.trigonometry;

import core.stdc.stdio;
import core.stdc.string: strlen, strcmp;

import punklabs.ui.label;
import punklabs.string.helper;

/// A simple knob base widget
class UIKnobBase : UIElement, IParameterListener
{
public:
nothrow:
@nogc:

	float sensitivity = 0.2f;

	this(UIContext context, Parameter param, uint flags = flagRaw, const(char)[] labelText=null, Font labelFont=null, bool showValue=true)
	{
		super(context, flags | flagAnimated);
		_param = cast(FloatParameter) param;
		_param.addListener(this);
		
		/*
			pointer, //default cursor
			linkSelect,
			move,
			drag,
			verticalResize,
			horizontalResize,
			diagonalResize,
			hidden
		*/
		//setCursorWhenMouseOver(MouseCursor.pointer);
		setCursorWhenDragged(MouseCursor.verticalResize);
		clearCrosspoints();
		
		// Use param.name()?
		// param.label() is the unit (ie. "Hz")
		if (labelText !is null && labelFont !is null) {
			_label = mallocNew!UILabel(context, labelFont, labelText);
			addChild(_label);
			if (showValue)
			{
				_labelValue = mallocNew!UILabel(context, labelFont, "");
				addChild(_labelValue);
				refreshLabelValue();
			}
		}
	}

	~this()
	{
		_param.removeListener(this);
	}

	override void onAnimate(double dt, double time)
	{
		if(atomicLoad(__parameterChanged))
		{
			atomicStore(__parameterChanged, false);
			refreshLabelValue();
		}
		updateLabel(); // Calling setX() on UILabel should only update if changed
	}

	RGBA getLabelColor()
	{
		return _labelColor;
	}
	RGBA getLabelBackgroundColor()
	{
		return _labelBackgroundColor;
	}

	RGBA getLabelValueColor()
	{
		return _labelValueColor;
	}
	
	override void reflow()
	{
		if (_label !is null) {
			updateLabel();
			auto p = position;
			auto x=p.min.x;
			auto y=p.min.y;
			auto w=p.size.x;
			auto h=p.size.y;
			auto labelHeight = min(w, h) * getLabelRelativeHeight();
			auto pos = rectangle(cast(int)(x+((w-w*getLabelRelativeWidth())/2)), cast(int)(y+h), cast(int)(w), cast(int)labelHeight);
			_label.position = pos;
			if (_labelValue !is null)
			{
				_labelValue.setColor(getLabelValueColor());
				auto lh = pos.size.y;
				pos.min.y += lh;
				pos.max.y += lh;
				_labelValue.position = pos;
			}
		}
	}

	void drawArc(ref Canvas canvas, float x, float y, float innerRadius, float outerRadius, float startAngle, float endAngle, bool anticlockwise)
	{
		drawArc(canvas, vec2f(x,y), innerRadius, outerRadius, startAngle, endAngle, anticlockwise);
	}
	void drawArc(ref Canvas canvas, vec2f center, float innerRadius, float outerRadius, float startAngle, float endAngle, bool anticlockwise)
	{
		canvas.arc(center, outerRadius, startAngle, endAngle, anticlockwise); 
		canvas.arc(center, innerRadius, endAngle, startAngle, !anticlockwise); 
	}

	void refreshLabelValue() {
		if (_labelValue !is null)
		{
			double normalizedValue = _param.getNormalized();
			auto suffix = _param.label();
			char[64] valueText;
			_param.stringFromNormalizedValue(normalizedValue, valueText.ptr, valueText.length);
			
			// Hack to get rid of ugly negative zero values!
			bool forceZero = (strcmp(valueText.ptr, "-0.0") == 0);
			char[64] fullLabelText;

			//snprintf(fullLabelText.ptr, fullLabelText.length, "%s%s", (forceZero ? "0.0".ptr : valueText.ptr), CString(suffix).storage);
			StringHelper.formatString(fullLabelText, "%s%s", (forceZero ? "0.0".ptr : valueText.ptr), CString(suffix).storage);
			
			//_param.stringFromNormalizedValue(normalizedValue, fullLabelText.ptr, 32);
			//_labelValue.setText(fullLabelText); // Sometimes leaves orphaned characters at end if sizes change... oops
			_labelValue.setText(fullLabelText[0..strlen(fullLabelText.ptr)]);
			_labelValue.setDirtyWhole();
		}
	}
	override void onBeginParameterEdit(Parameter sender)
	{
		setDirtyWhole();
	}
	override void onParameterChanged(Parameter sender)
	{
		atomicStore(__parameterChanged, true);
		setDirtyWhole();
	}
	override void onEndParameterEdit(Parameter sender)
	{
		setDirtyWhole();
	}

	override bool onMouseClick(int x, int y, int button, bool isDoubleClick, MouseState mstate)
	{
		// double-click => set to default
		if (isDoubleClick || mstate.altPressed)
		{
			_param.beginParamEdit();
			_param.setFromGUI(_param.defaultValue());
			_param.endParamEdit();
		}
		return true; // to initiate dragging
	}

	override void onBeginDrag()
	{
		_param.beginParamEdit();
	}
	override void onStopDrag()
	{
		_param.endParamEdit();
	}
	override void onMouseDrag(int x, int y, int dx, int dy, MouseState mstate)
	{
		// FUTURE: replace by actual trail height instead of total height
		float displacementInHeight = cast(float)(dy) / _position.height;
		
		float modifier = 1.0f;
		if (mstate.shiftPressed || mstate.ctrlPressed) modifier *= 0.1f;
		
		double oldParamValue = _param.getNormalized();
		double newParamValue = oldParamValue - displacementInHeight * modifier * sensitivity;
		if (mstate.altPressed)
			newParamValue = _param.getNormalizedDefault();
		
		if (y > _mousePosOnLast0Cross) return;
		if (y < _mousePosOnLast1Cross) return;
		if (newParamValue <= 0 && oldParamValue > 0) _mousePosOnLast0Cross = y;
		if (newParamValue >= 1 && oldParamValue < 1) _mousePosOnLast1Cross = y;
		if (newParamValue < 0) newParamValue = 0;
		if (newParamValue > 1) newParamValue = 1;
		if (newParamValue > 0) _mousePosOnLast0Cross = float.infinity;
		if (newParamValue <1) _mousePosOnLast1Cross = -float.infinity;
		
		if (newParamValue != oldParamValue)
		{
			if (auto p = cast(FloatParameter)_param)
			{
				p.setFromGUINormalized(newParamValue);
			}
			else
				assert(false); // only float parameters supported
		}
	}

	override void onMouseEnter()
	{
		setDirtyWhole();
	}

	override void onMouseExit()
	{
		setDirtyWhole();
	}
	
private:
    shared(bool) __parameterChanged = false;  // access to this is through atomic ops

	void updateLabel()
	{
		if (_label !is null) {  // Calling setX() on UILabel should only update if changed
			_label.setColor(getLabelColor());
			_label.setBackgroundColor(getLabelBackgroundColor());
		}
	}
	float getLabelRelativeWidth()
	{
		return 1.0f;
	}
	float getLabelRelativeHeight()
	{
		return 0.2f;
	}
	FloatParameter _param = null;
	
	UILabel _label = null;
	RGBA _labelColor = RGBA(255, 255, 255, 255);
	RGBA _labelBackgroundColor = RGBA(255, 255, 255, 255);

	UILabel _labelValue = null;
	RGBA _labelValueColor = RGBA(255, 255, 255, 255);
	
	float _mousePosOnLast0Cross;
	float _mousePosOnLast1Cross;
	
	void clearCrosspoints()
	{
		_mousePosOnLast0Cross = float.infinity;
		_mousePosOnLast1Cross = -float.infinity;
	}
}
